'2008 IDesign Inc. 
'Questions? Comments? go to 
'http://www.idesign.net


Imports Microsoft.VisualBasic
Imports System
Imports System.Transactions
Imports System.Diagnostics
Imports System.Collections.Generic

Namespace ServiceModelEx
   Public Class Transactional(Of T)
	   Implements ISinglePhaseNotification
	  Private m_Value As T
	  Private m_TemporaryValue As T
	  Private m_CurrentTransaction As Transaction
	  Private m_Lock As TransactionalLock

        Public Sub New(ByVal value As T)
            CreateDefaults(value)
        End Sub

        Private Sub CreateDefaults(ByVal value As T)
            m_Lock = New TransactionalLock()
            m_Value = value
        End Sub

        Public Sub New(ByVal transactional As Transactional(Of T))
            Me.New(transactional.Value)
        End Sub
	  Public Sub New()
            CreateDefaults(Nothing)
	  End Sub
	  Shared Sub New()
		 ResouceManager.ConstrainType(GetType(T))
	  End Sub
	  Private Sub SinglePhaseCommit(ByVal singlePhaseEnlistment As SinglePhaseEnlistment) Implements ISinglePhaseNotification.SinglePhaseCommit
		 Commit()
		 singlePhaseEnlistment.Committed()
	  End Sub
	  Private Sub Commit()
		 Dim disposable As IDisposable = TryCast(m_Value, IDisposable)
		 If disposable IsNot Nothing Then
			disposable.Dispose()
		 End If
		 m_Value = m_TemporaryValue
		 m_CurrentTransaction = Nothing
		 m_TemporaryValue= Nothing
		 m_Lock.Unlock()
	  End Sub
	  Private Sub Commit(ByVal enlistment As Enlistment) Implements IEnlistmentNotification.Commit
		 Commit()
		 enlistment.Done()
	  End Sub

	  Private Sub InDoubt(ByVal enlistment As Enlistment) Implements IEnlistmentNotification.InDoubt
		 m_Lock.Unlock()
		 enlistment.Done()
	  End Sub
	  Private Sub Prepare(ByVal preparingEnlistment As PreparingEnlistment) Implements IEnlistmentNotification.Prepare
		 preparingEnlistment.Prepared()
	  End Sub

	  Private Sub Rollback(ByVal enlistment As Enlistment) Implements IEnlistmentNotification.Rollback
		 m_CurrentTransaction = Nothing

		 Dim disposable As IDisposable = TryCast(m_TemporaryValue, IDisposable)
		 If disposable IsNot Nothing Then
			disposable.Dispose()
		 End If

		 m_TemporaryValue = Nothing
		 m_Lock.Unlock()
		 enlistment.Done()
	  End Sub
	  Private Sub Enlist(ByVal t As T)
		 Debug.Assert(m_CurrentTransaction Is Nothing)
		 m_CurrentTransaction = Transaction.Current
		 Debug.Assert(m_CurrentTransaction.TransactionInformation.Status = TransactionStatus.Active)
		 m_CurrentTransaction.EnlistVolatile(Me,EnlistmentOptions.None)
		 m_TemporaryValue = ResouceManager.Clone(t)
	  End Sub
	  Private Sub SetValue(ByVal t As T)
		 m_Lock.Lock()
		 If m_CurrentTransaction Is Nothing Then
			If Transaction.Current Is Nothing Then
			   m_Value = t
			   Return
			Else
			   Enlist(t)
			   Return
			End If
		 Else
			'Must have acquired the lock
			Debug.Assert(m_CurrentTransaction = Transaction.Current,"Invalid state in the volatile resource state machine")
			m_TemporaryValue = t
		 End If
	  End Sub
	  Private Function GetValue() As T
		 m_Lock.Lock()
		 If m_CurrentTransaction Is Nothing Then
			If Transaction.Current Is Nothing Then
			   Return m_Value
			Else
			   Enlist(m_Value)
			End If
		 End If
		 'Must have acquired the lock
		 Debug.Assert(m_CurrentTransaction = Transaction.Current,"Invalid state in the volatile resource state machine")

		 Return m_TemporaryValue
	  End Function
	  Public Property Value() As T
		 Get
			Return GetValue()
		 End Get
		 Set(ByVal value As T)
			SetValue(value)
		 End Set
	  End Property
	  Public Shared Widening Operator CType(ByVal transactional As Transactional(Of T)) As T
		 Return transactional.Value
	  End Operator
	  Public Shared Operator =(ByVal t1 As Transactional(Of T), ByVal t2 As Transactional(Of T)) As Boolean
		 ' Is t1 and t2 null (check the value as well).
		 Dim t1Null As Boolean = (Object.ReferenceEquals(t1,Nothing) OrElse t1.Value Is Nothing)
		 Dim t2Null As Boolean = (Object.ReferenceEquals(t2,Nothing) OrElse t2.Value Is Nothing)

		 ' If they are both null,return true.
		 If t1Null AndAlso t2Null Then
			Return True
		 End If

		 ' If one is null,return false.
		 If t1Null OrElse t2Null Then
			Return False
		 End If
		 Return EqualityComparer(Of T).Default.Equals(t1.Value,t2.Value)
	  End Operator
	  Public Shared Operator =(ByVal t1 As Transactional(Of T), ByVal t2 As T) As Boolean
		 ' Is t1 and t2 null (check the value as well).
		 Dim t1Null As Boolean = (Object.ReferenceEquals(t1,Nothing) OrElse t1.Value Is Nothing)
		 Dim t2Null As Boolean = t2 Is Nothing

		 ' If they are both null,return true.
		 If t1Null AndAlso t2Null Then
			Return True
		 End If

		 ' If one is null,return false.
		 If t1Null OrElse t2Null Then
			Return False
		 End If
		 Return EqualityComparer(Of T).Default.Equals(t1.Value,t2)
	  End Operator
	  Public Shared Operator =(ByVal t1 As T, ByVal t2 As Transactional(Of T)) As Boolean
		 ' Is t1 and t2 null (check the value as well)
		 Dim t1Null As Boolean = t1 Is Nothing
		 Dim t2Null As Boolean = (Object.ReferenceEquals(t2,Nothing) OrElse t2.Value Is Nothing)

		 ' If they are both null,return true.
		 If t1Null AndAlso t2Null Then
			Return True
		 End If

		 ' If one is null,return false.
		 If t1Null OrElse t2Null Then
			Return False
		 End If
		 Return EqualityComparer(Of T).Default.Equals(t1,t2.Value)
	  End Operator
	  Public Shared Operator <>(ByVal t1 As T, ByVal t2 As Transactional(Of T)) As Boolean
            Return Not (t1 = t2)
	  End Operator
	  Public Shared Operator <>(ByVal t1 As Transactional(Of T), ByVal t2 As T) As Boolean
            Return Not (t1 = t2)
	  End Operator
	  Public Shared Operator <>(ByVal t1 As Transactional(Of T), ByVal t2 As Transactional(Of T)) As Boolean
		 Return Not(t1 = t2)
	  End Operator
	  Public Overrides Function GetHashCode() As Integer
		 Return Value.GetHashCode()
	  End Function
	  Public Overrides Overloads Function Equals(ByVal obj As Object) As Boolean
		 Return Value.Equals(obj)
	  End Function
   End Class
End Namespace
