'2008 IDesign Inc. 
'Questions? Comments? go to 
'http://www.idesign.net


Imports Microsoft.VisualBasic
Imports System
Imports System.Threading
Imports System.Collections.Generic
Imports System.Transactions
Imports System.Diagnostics

Namespace ServiceModelEx
   ''' <summary>
   ''' Protects a resource manager by proving exclusive transaction isolation (Serializable level only)
   ''' </summary>
   Public Class TransactionalLock
	  'Transaction that tried to acquire the lock while the lock is owned by another transaction are placed in this queue
	  Private m_PendingTransactions As LinkedList(Of KeyValuePair(Of Transaction,ManualResetEvent)) = New LinkedList(Of KeyValuePair(Of Transaction,ManualResetEvent))()
        Private m_OwningTransaction As Transaction
        Private manualEvent As ManualResetEvent
        Private pair As KeyValuePair(Of Transaction, ManualResetEvent)

        Private Property OwningTransaction() As Transaction
            Get
                SyncLock Me
                    Return m_OwningTransaction
                End SyncLock
            End Get
            Set(ByVal value As Transaction)
                SyncLock Me
                    m_OwningTransaction = value
                End SyncLock
            End Set
        End Property
        Public ReadOnly Property Locked() As Boolean
            Get
                Return OwningTransaction IsNot Nothing
            End Get
        End Property
        ''' <summary>
        ''' Acquires the lock for the exclusive use of a transaction. If another transaction owns the lock,it blocks the calling transaction and places it in a queue. If the transaction owns the lock already Lock() does nothing. 
        ''' </summary>
        Public Sub Lock()
            Lock(Transaction.Current)
        End Sub
        Private Sub Lock(ByVal transaction As Transaction)
            Monitor.Enter(Me)
            If OwningTransaction Is Nothing Then
                If transaction Is Nothing Then
                    Monitor.Exit(Me)
                    Return
                Else
                    Debug.Assert(transaction.IsolationLevel = IsolationLevel.Serializable)

                    'Acquire the lock
                    OwningTransaction = transaction
                    Monitor.Exit(Me)
                    Return
                End If
            Else 'Some transaction owns the lock
                'Is it the same one?
                If OwningTransaction = transaction Then
                    Monitor.Exit(Me)
                    Return
                Else 'Need to lock

                    manualEvent = New ManualResetEvent(False)

                    Me.pair = New KeyValuePair(Of Transaction, ManualResetEvent)(transaction, Me.manualEvent)
                    m_PendingTransactions.AddLast(Me.pair)

                    If transaction IsNot Nothing Then
                        Debug.Assert(transaction.TransactionInformation.Status = TransactionStatus.Active)

                        AddHandler transaction.TransactionCompleted, AddressOf OnTransactionCompleted
                    End If
                    Monitor.Exit(Me)
                    'Block the transaction or the calling thread
                    Me.manualEvent.WaitOne()
                    SyncLock Me.manualEvent 'To deal with race condition of the other threads setting the handle
                        Me.manualEvent.Close()
                    End SyncLock
                End If
            End If
        End Sub

        Private Sub OnTransactionCompleted(ByVal sender As Object, ByVal e As EventArgs)
            'Since the transaction can abort or just time out while blocking,unblock it when it is completed and remove from the queue
            SyncLock Me
                m_PendingTransactions.Remove(pair)
            End SyncLock
            SyncLock Me.manualEvent 'To deal with race condition of the handle closed between the check and the set
                If Me.manualEvent.SafeWaitHandle.IsClosed = False Then
                    Me.manualEvent.Set()
                End If
            End SyncLock
        End Sub
	  'Releases the transaction lock and allows the next pending transaction to quire it. 
	  Public Sub Unlock()
		 Debug.Assert(Locked)
		 SyncLock Me
			OwningTransaction = Nothing

			Dim node As LinkedListNode(Of KeyValuePair(Of Transaction,ManualResetEvent)) = Nothing

			If m_PendingTransactions.Count > 0 Then
			   node = m_PendingTransactions.First
			   m_PendingTransactions.RemoveFirst()
			End If
			If node IsNot Nothing Then
			   Dim transaction As Transaction = node.Value.Key
			   Dim manualEvent As ManualResetEvent = node.Value.Value
			   Lock(transaction)
			   SyncLock manualEvent 'To deal with race condition of the handle closed between the check and the set
				  If manualEvent.SafeWaitHandle.IsClosed = False Then
					 manualEvent.Set()
				  End If
			   End SyncLock
			End If
		 End SyncLock
	  End Sub
   End Class
End Namespace
